"""
Module to easily convert between Mach number, true airspeed (TAS), calibrated
airspeed (CAS), and equivalent airspeed (EAS) for given altitude(s).
Additional flight condition data and atmospheric data is computed.
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


Author: Matthew C. Jones
Email: matt.c.jones.aoe@gmail.com

:copyright: 2021 Matthew C. Jones
:license: MIT License, see LICENSE for more details.
"""

from .flightcondition import FlightCondition
